/*
 * LEDD Project
 * Copyright (C) 2015 LEDD Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.idlegandalf.ledd;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.graphics.Color;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.support.annotation.Nullable;
import android.util.SparseArray;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.CompoundButton;
import android.widget.LinearLayout;

import com.google.android.material.snackbar.Snackbar;
import com.google.common.reflect.TypeToken;
import com.google.gson.Gson;
import com.idlegandalf.ledd.callbacks.ReceiveColorCallback;
import com.idlegandalf.ledd.callbacks.StripesCallback;
import com.idlegandalf.ledd.components.HSV;
import com.idlegandalf.ledd.components.LedDDaemon;
import com.idlegandalf.ledd.components.LedStripe;
import com.idlegandalf.ledd.components.Profile;
import com.idlegandalf.ledd.fragments.AddProfileDialog;
import com.idlegandalf.ledd.fragments.AddStripeDialog;
import com.idlegandalf.ledd.helper.LedDHelper;
import com.idlegandalf.ledd.utils.RateLimiter;
import com.madrapps.pikolo.HSLColorPicker;
import com.madrapps.pikolo.listeners.SimpleColorSelectionListener;
import com.mikepenz.materialdrawer.Drawer;
import com.mikepenz.materialdrawer.DrawerBuilder;
import com.mikepenz.materialdrawer.holder.BadgeStyle;
import com.mikepenz.materialdrawer.interfaces.OnCheckedChangeListener;
import com.mikepenz.materialdrawer.model.DividerDrawerItem;
import com.mikepenz.materialdrawer.model.PrimaryDrawerItem;
import com.mikepenz.materialdrawer.model.SecondaryDrawerItem;
import com.mikepenz.materialdrawer.model.SecondarySwitchDrawerItem;
import com.mikepenz.materialdrawer.model.SwitchDrawerItem;
import com.mikepenz.materialdrawer.model.interfaces.IDrawerItem;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;

import androidx.appcompat.app.AppCompatActivity;
import androidx.appcompat.widget.Toolbar;
import butterknife.BindView;
import butterknife.ButterKnife;
import timber.log.Timber;

public class ColorActivity extends AppCompatActivity implements Drawer.OnDrawerItemClickListener, Drawer.OnDrawerItemLongClickListener, OnCheckedChangeListener {

    @BindView(R.id.main_layout)
    LinearLayout scrollView;
    @BindView(R.id.toolbar)
    Toolbar toolbar;
    @BindView(R.id.picker)
    HSLColorPicker colorPicker;
    Drawer mDrawer;
    private List<IDrawerItem> mDaemons;
    private refreshDaemonsListener daemonsListener;
    private List<LedStripe> ledStripes;
    private List<Profile> profiles;
    private LedStripe mCurrentStripe;
    private boolean fromOnCreate = true;
    private boolean autoColorSet = false;
    private RateLimiter limiter;
    private boolean isRefreshRunning = false;
    private Type profileList = new TypeToken<List<Profile>>() {
    }.getType();
    private PrimaryDrawerItem menuProfiles;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_color);
        ButterKnife.bind(this);

        if (BuildConfig.DEBUG) {
            Timber.plant(new Timber.DebugTree());
        }

        // check for connectivity
        ConnectivityManager connMgr = (ConnectivityManager) getSystemService(Context.CONNECTIVITY_SERVICE);
        NetworkInfo networkInfo = connMgr.getActiveNetworkInfo();
        if (networkInfo == null || !networkInfo.isConnected() || networkInfo.getType() != ConnectivityManager.TYPE_WIFI) {
            Snackbar.make(scrollView, R.string.snackbar_no_wlan, Snackbar.LENGTH_INDEFINITE).setAction("RETRY", v -> {
                refreshStripes();
            }).show();
        }

        mDaemons = new ArrayList<>();
        limiter = new RateLimiter(30.0, 1000);

        colorPicker.setColorSelectionListener(new SimpleColorSelectionListener() {
            @Override
            public void onColorSelected(int color) {
                if (mCurrentStripe != null && limiter.check() && !autoColorSet) {
                    mCurrentStripe.setColor(color);
                }
            }
        });

        toolbar.setTitle(R.string.app_name);

        mDrawer = new DrawerBuilder()
                .withActivity(this)
                .withToolbar(toolbar)
                .withOnDrawerItemClickListener(this)
                .withOnDrawerItemLongClickListener(this)
                .addStickyDrawerItems(
                        new PrimaryDrawerItem().withName(R.string.text_add_stripe).withTag("add_stripe").withSelectable(false).withIcon(R.drawable.ic_add_circle_black_48dp),
                        new PrimaryDrawerItem().withName(R.string.text_settings).withTag("settings").withSelectable(false).withIcon(R.drawable.ic_tune_black_48dp)
                )
                .addDrawerItems(
                        new DividerDrawerItem(),
                        menuProfiles = new PrimaryDrawerItem()
                                .withName("Profiles")
                                .withIcon(R.drawable.ic_save_black_48dp)
                                .withSelectable(false)
                                .withIsExpanded(true)
                                .withBadgeStyle(new BadgeStyle().withColor(Color.CYAN).withTextColor(Color.WHITE))
                )
                .build();

        menuProfiles.withSubItems(new SecondaryDrawerItem().withName("Add Profile").withTag("add_profile").withSelectable(false).withIcon(R.drawable.ic_add_circle_outline_black_48dp));

        // enable HomeButton navigation to drawer
        if (getSupportActionBar() != null) {
            getSupportActionBar().setDisplayHomeAsUpEnabled(false);
            mDrawer.getActionBarDrawerToggle().setDrawerIndicatorEnabled(true);
        }

        List<IDrawerItem> subProfiles = menuProfiles.getSubItems();
        if (PreferenceManager.getDefaultSharedPreferences(this).contains("profiles")) {
            profiles = new Gson().fromJson(PreferenceManager.getDefaultSharedPreferences(this).getString("profiles", ""), profileList);

            for (Profile profile : profiles) {
                subProfiles.add(0, new SecondaryDrawerItem().withName(profile.getName()).withTag(profile.getValues()).withSelectable(false).withIcon(R.drawable.ic_note_black_48dp));
            }
        } else {
            profiles = new ArrayList<>();
        }

        menuProfiles.withSubItems(subProfiles);
        mDrawer.updateItem(menuProfiles);

        daemonsListener = new refreshDaemonsListener();
        registerReceiver(daemonsListener, new IntentFilter(ColorApplication.INTENT_ACTION_REFRESH));

    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        getMenuInflater().inflate(R.menu.color, menu);

        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        return super.onOptionsItemSelected(item);
    }

    @Override
    protected void onRestart() {
        super.onRestart();
        ColorApplication.getInstance().onResume();
        registerReceiver(daemonsListener, new IntentFilter(ColorApplication.INTENT_ACTION_REFRESH));
        refreshStripes();
    }

    @Override
    public void onCheckedChanged(IDrawerItem drawerItem, CompoundButton buttonView, boolean isChecked) {
        LedStripe stripe = null;
        for (LedStripe ledStripe : ledStripes) {
            if (((int) drawerItem.getTag()) == ledStripe.getId()) {
                stripe = ledStripe;
                break;
            }
        }

        if (stripe != null) {
            if (isChecked) {
                if (stripe == mCurrentStripe) {
                    if (colorPicker.getSolidColor() != Color.BLACK)
                        stripe.setColor(colorPicker.getSolidColor());
                    else {
                        colorPicker.setColor(Color.WHITE);
                    }
                } else {
                    stripe.setColor(Color.WHITE);
                }
            } else {
                if (stripe == mCurrentStripe)
                    colorPicker.setColor(Color.BLACK);
                else
                    stripe.setColor(Color.BLACK);
            }
        }
    }

    @Override
    public boolean onKeyDown(int keyCode, KeyEvent event) {
        if (keyCode == KeyEvent.KEYCODE_MENU) {
            if (mDrawer.isDrawerOpen()) {
                mDrawer.closeDrawer();
            } else {
                mDrawer.openDrawer();
            }

            return true;
        }
        return super.onKeyDown(keyCode, event);
    }

    @Override
    protected void onPause() {
        super.onPause();
        ColorApplication.getInstance().teardown();
        unregisterReceiver(daemonsListener);

        if (mCurrentStripe != null)
            PreferenceManager.getDefaultSharedPreferences(this).edit().putInt("lastStripe", mCurrentStripe.getId()).apply();

        if (profiles != null)
            PreferenceManager.getDefaultSharedPreferences(this).edit().putString("profiles", new Gson().toJson(profiles, profileList)).apply();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        ColorApplication.getInstance().teardown();
    }

    @Override
    protected void onResume() {
        super.onResume();

        if (!fromOnCreate) {
            ColorApplication.getInstance().onResume();
            registerReceiver(daemonsListener, new IntentFilter(ColorApplication.INTENT_ACTION_REFRESH));
            refreshStripes();
        } else {
            fromOnCreate = false;
        }
    }


    @SuppressWarnings("unchecked")
    @Override
    public boolean onItemClick(View view, int position, IDrawerItem drawerItem) {
        if (drawerItem.getTag() instanceof String) {
            String tag = (String) drawerItem.getTag();

            switch (tag) {
                case "add_stripe":
                    new AddStripeDialog().show(getFragmentManager(), "");
                    return true;
                case "settings":
                    return true;
                case "add_profile":
                    new AddProfileDialog().show(getFragmentManager(), "");
                    return true;
            }
        }

        if (drawerItem.getTag() instanceof SparseArray) {
            SparseArray<HSV> sparseArray = (SparseArray<HSV>) drawerItem.getTag();

            if (ledStripes != null) {
                for (LedStripe ledStripe : ledStripes) {
                    String json_workaround = new Gson().toJson(sparseArray.get(ledStripe.getId()));
                    HSV nHSV = new Gson().fromJson(json_workaround, new TypeToken<HSV>() {
                    }.getType());

                    if (nHSV != null) {
                        ledStripe.setColor(nHSV);
                    }
                }
            }
        }

        if (drawerItem.isSelectable()) {
            if (drawerItem.getTag() instanceof Integer) {
                int tag = (int) drawerItem.getTag();

                for (LedStripe stripe : ledStripes) {
                    if (tag == stripe.getId()) {
                        selectStripe(stripe);
                        return true;
                    }
                }
            }
        }
        return false;
    }

    @Override
    public boolean onItemLongClick(View view, int position, IDrawerItem drawerItem) {
        Timber.d("Longclick of %s", drawerItem.getClass().getName());

        return false;
    }

    public List<LedStripe> getStripes() {
        return ledStripes;
    }

    public void addProfile(Profile profile) {
        if (profiles != null) {
            profiles.add(profile);

            List<IDrawerItem> subProfiles = menuProfiles.getSubItems();
            subProfiles.add(0, new SecondaryDrawerItem().withName(profile.getName()).withSelectable(false).withTag(profile.getValues()).withIcon(R.drawable.ic_note_black_48dp));

            menuProfiles.withSubItems(subProfiles).withIsExpanded(true).withBadge(String.valueOf(profiles.size()));
            mDrawer.updateItem(menuProfiles);
        }
    }

    @SuppressWarnings("deprecated")
    public void refreshStripes() {
        if (isRefreshRunning)
            return;
        isRefreshRunning = true;

        int i = 1;
        for (final LedDDaemon dDaemon : ColorApplication.getInstance().getDaemons()) {
            LedDHelper helper = ColorApplication.getInstance().getHelperForDaemon(dDaemon);

            if (helper != null) {
                final int finalI = i;
                helper.getStripes(new StripesCallback() {
                    @Override
                    public void onSuccess(List<LedStripe> stripes) {
                        ledStripes = stripes;
                        runOnUiThread(() -> {
                            PrimaryDrawerItem nDaemonMenu = null;
                            long selectedItemId = -1;
                            for (IDrawerItem drawerItem : mDaemons) {
                                if (((PrimaryDrawerItem) drawerItem).getName().getText().equals(dDaemon.toString())) {
                                    nDaemonMenu = (PrimaryDrawerItem) drawerItem;
                                    break;
                                }
                            }

                            if (nDaemonMenu == null) {
                                nDaemonMenu = new PrimaryDrawerItem().withName(dDaemon.toString()).withTag(finalI).withSelectable(false).withIsExpanded(true).withIcon(R.drawable.ic_computer_black_48dp);
                                mDrawer.addItemAtPosition(nDaemonMenu, 0);
                                mDaemons.add(nDaemonMenu);
                            }

                            int lastStripeId = -1;
                            if (PreferenceManager.getDefaultSharedPreferences(getApplicationContext()).contains("lastStripe")) {
                                lastStripeId = PreferenceManager.getDefaultSharedPreferences(getApplicationContext()).getInt("lastStripe", -1);
                            }

                            List<IDrawerItem> mStripes = new ArrayList<>();
                            for (LedStripe stripe : ledStripes) {
                                final SwitchDrawerItem sItem = new SwitchDrawerItem()
                                        .withName(stripe.getName())
                                        .withTag(stripe.getId())
                                        .withIcon(R.drawable.ic_wb_iridescent_black_48dp)
                                        .withOnCheckedChangeListener(ColorActivity.this);

                                mStripes.add(sItem);

                                if (lastStripeId != -1 && stripe.getId() == lastStripeId) {
                                    selectStripe(stripe);
                                    selectedItemId = sItem.getIdentifier();
                                }
                            }

                            nDaemonMenu.withSubItems(mStripes);
                            nDaemonMenu.withBadge(String.valueOf(mStripes.size()));
                            //noinspection deprecation
                            nDaemonMenu.withBadgeStyle(new BadgeStyle().withColor(getResources().getColor(R.color.material_drawer_accent)).withTextColor(Color.WHITE));
                            mDrawer.updateItem(nDaemonMenu);
                            if (selectedItemId != -1)
                                mDrawer.setSelection(selectedItemId);

                            //noinspection ConstantConditions
                            Snackbar.make(findViewById(android.R.id.content), R.string.snackbar_stripes_reloaded, Snackbar.LENGTH_LONG).show();
                        });


                        for (LedStripe stripe : ledStripes) {
                            stripe.getColor(new ReceiveColorCallback() {
                                @Override
                                public void onConnectionFailed(String message) {

                                }

                                @Override
                                public void onColorReceived(LedStripe stripe) {
                                    if (stripe.isOn()) {
                                        IDrawerItem sItem = findItemForStripe(stripe);

                                        if (sItem != null) {
                                            final SecondarySwitchDrawerItem cItem = (SecondarySwitchDrawerItem) sItem;
                                            cItem.withSwitchEnabled(true);

                                            runOnUiThread(() -> mDrawer.updateItem(cItem));
                                        }
                                    }
                                }

                                @Override
                                public void onReceiveFailed(int code, String msg) {

                                }

                            });
                        }
                    }

                    @Override
                    public void onGetFailed(int code, String message) {
                        //noinspection ConstantConditions
                        Snackbar.make(findViewById(android.R.id.content), getString(R.string.snackbar_no_connection_stripes) + message, Snackbar
                                .LENGTH_LONG).show();

                    }

                    @Override
                    public void onConnectionFailed(String message) {
                        //noinspection ConstantConditions
                        Snackbar.make(findViewById(android.R.id.content), getString(R.string.snackbar_connection_failed) + message, Snackbar
                                .LENGTH_LONG).show();
                    }
                });

            }

            i++;
        }

        isRefreshRunning = false;
    }

    private void selectStripe(LedStripe stripe) {
        mCurrentStripe = stripe;

        toolbar.setTitle(stripe.getName());
        autoColorSet = true;
        mDrawer.closeDrawer();
        mCurrentStripe.getColor(new ReceiveColorCallback() {
            @Override
            public void onColorReceived(final LedStripe stripe) {
                final HSV cColor = stripe.getColor();
                final int color = Color.HSVToColor(new float[]{(float) cColor.getHue(), (float) cColor.getSaturation(), (float) cColor.getValue()});

                runOnUiThread(() -> {
                    colorPicker.setColor(color);
                    autoColorSet = false;

                    if (stripe.isOn()) {
                        IDrawerItem fItem = findItemForStripe(stripe);

                        if (fItem != null)
                            ((SecondarySwitchDrawerItem) fItem).withSwitchEnabled(true);
                    }
                });
            }

            @Override
            public void onReceiveFailed(int code, String msg) {
                autoColorSet = false;
            }

            @Override
            public void onConnectionFailed(String message) {
                autoColorSet = false;
            }
        });
    }

    @Nullable
    private IDrawerItem findItemForStripe(LedStripe stripe) {
        for (IDrawerItem dItem : mDaemons) {
            if (((PrimaryDrawerItem) dItem).getSubItems() != null) {
                for (IDrawerItem sItem : ((PrimaryDrawerItem) dItem).getSubItems()) {
                    if (((int) sItem.getTag()) == stripe.getId()) {
                        return sItem;
                    }
                }
            }
        }

        return null;
    }

    protected class refreshDaemonsListener extends BroadcastReceiver {

        @Override
        public void onReceive(Context context, Intent intent) {
            if (intent.getAction().equals(ColorApplication.INTENT_ACTION_REFRESH)) {
                refreshStripes();
            }
        }
    }
}
